/*-----------------------------------------------------------------------*/

#include <stdint.h>

#include "diskio.h"

typedef volatile uint8_t vuint8_t;
typedef volatile uint16_t vuint16_t;
typedef volatile uint32_t vuint32_t;

/* both ED64 and N64 Myth HW are at the same address */
vuint32_t *regs_ptr = (vuint32_t *)0xA8040000;

/* ED64 support */

vuint8_t spi_cfg = 0;
vuint8_t evd_cfg = 0;

uint16_t evd_firm;

#define REG_CFG 			0
#define REG_STATUS 			1
#define REG_DMA_LEN 		2
#define REG_DMA_RAM_ADDR 	3
#define REG_MSG 			4
#define REG_DMA_CFG 		5
#define REG_SPI 			6
#define REG_SPI_CFG 		7
#define REG_KEY 			8
#define REG_SAV_CFG 		9
#define REG_SEC 			10
#define REG_VER 			11

#define REG_CFG_CNT 		16
#define REG_CFG_DAT 		17
#define REG_MAX_MSG 		18
#define REG_CRC 			19

#define DCFG_SD_TO_RAM 		1
#define DCFG_RAM_TO_SD 		2
#define DCFG_FIFO_TO_RAM 	3
#define DCFG_RAM_TO_FIFO 	4

#define ED_CFG_SDRAM_ON 	0
#define ED_CFG_SWAP 		1
#define ED_CFG_WR_MOD 		2
#define ED_CFG_WR_ADDR_MASK 3

#define SAV_EEP_ON 			0
#define SAV_SRM_ON 			1
#define SAV_EEP_SIZE 		2
#define SAV_SRM_SIZE 		3

#define SPI_CFG_SPD0 		0
#define SPI_CFG_SPD1 		1
#define SPI_CFG_SS 			2
#define SPI_CFG_RD 			3
#define SPI_CFG_DAT 		4
#define SPI_CFG_1BIT 		5

uint16_t evd_readReg(uint8_t reg)
{
    vuint32_t tmp = regs_ptr[0];

    return regs_ptr[reg];
}

void evd_init()
{
    vuint8_t val;

    val = regs_ptr[0];

    spi_cfg = (0 << SPI_CFG_SPD0) | (1 << SPI_CFG_SPD1) | (1 << SPI_CFG_SS);
    evd_cfg = (1 << ED_CFG_SDRAM_ON);

    val = regs_ptr[0];
    regs_ptr[REG_KEY] = 0x1234;

    val = regs_ptr[0];
    regs_ptr[REG_CFG] = evd_cfg;
    val = regs_ptr[0];
    regs_ptr[REG_SPI_CFG] = spi_cfg;

    evd_firm = evd_readReg(REG_VER);
}

/* N64 Myth support */

uint16_t neo_firm;

#define SAVE_IO   0
#define CIC_IO    2
#define ROM_BANK  6
#define ROM_SIZE  4
#define RUN_IO    16
#define INT_IO    18
#define NEO_IO    20
#define SRAM2C_IO 22
#define ROMC_IO   24
#define ROMSW_IO  26
#define SRAM2C_I0 27
#define RST_IO    28
#define CIC_EN    30
#define CPID_IO   32

#define _neo_asic_op(cmd) *(vuint32_t *)(0xB2000000 | (cmd<<1))
#define _neo_asic_wop(cmd) *(vuint16_t *)(0xB2000000 | (cmd<<1))

void neo_sync_bus(void)
{
    asm volatile
    (
        ".set   push\n"
        ".set   noreorder\n"
        "lui    $t0,0xb000\n"
        "lw     $zero,($t0)\n"
        "jr     $ra\n"
        "nop\n"
        ".set pop\n"
    );
}

/* do a Neo Flash ASIC command */
static uint16_t _neo_asic_cmd(uint32_t cmd, uint32_t full)
{
    uint32_t data;

    if (full)
    {
        /* need to switch rom bus for ASIC operations */
        regs_ptr[INT_IO] = 0xFFFFFFFF;  	/* enable multi-card mode */
        neo_sync_bus();
        regs_ptr[ROMSW_IO] = 0x00000000;    /* gba mapped to 0xB2000000 - 0xB3FFFFFF */
        neo_sync_bus();
        regs_ptr[ROM_BANK] = 0x00000000;
        neo_sync_bus();
        regs_ptr[ROM_SIZE] = 0x000C000C;    /* bank size = 256Mbit */
        neo_sync_bus();
    }

    /* at this point, assumes rom bus in a mode that allows ASIC operations */

    regs_ptr[NEO_IO] = 0xFFFFFFFF;          /* 16 bit mode */
    neo_sync_bus();

    /* do unlocking sequence */
    _neo_asic_op(0x00FFD200);
    _neo_asic_op(0x00001500);
    _neo_asic_op(0x0001D200);
    _neo_asic_op(0x00021500);
    _neo_asic_op(0x00FE1500);
    /* do ASIC command */
    data = _neo_asic_op(cmd);

    regs_ptr[NEO_IO] = 0x00000000;          /* 32 bit mode */
    neo_sync_bus();

    return (uint16_t)(data & 0x0000FFFF);
}

static uint32_t neo_get_cpld(void)
{
    uint32_t data;

    regs_ptr[INT_IO] = 0xFFFFFFFF;          /* enable multi-card mode */
    neo_sync_bus();

    regs_ptr[CPID_IO] = 0xFFFFFFFF;         /* enable CPLD ID read */
    neo_sync_bus();
    data = *(vuint32_t *)0xB0000000;
    neo_sync_bus();
    regs_ptr[CPID_IO] = 0x00000000;         /* disable CPLD ID read */
    neo_sync_bus();

    return data;
}

static void neo_enable_sd(void)
{
    /* Neo2-SD (compatible with Neo2-Pro for 16MB of psram) */
    _neo_asic_cmd(0x00E21500, 1);           /* GBA CARD WE ON ! */
    _neo_asic_cmd(0x00372682, 0);           /* set cr = game flash and write enabled, enable SD interface */
    _neo_asic_cmd(0x00DAAF4E, 0);           /* select psram */
    _neo_asic_cmd(0x00EE0630, 0);           /* set cr1 = enable extended address bus */

    regs_ptr[ROMSW_IO] = 0xFFFFFFFF;        /* gba mapped to 0xB0000000 - 0xB3FFFFFF */
    neo_sync_bus();
    regs_ptr[ROM_SIZE] = 0x000C000C;        /* bank size = 256Mbit (128Mbit psram : 128Mbit SD iface) */
    neo_sync_bus();
}

/*-----------------------------------------------------------------------*/

#define SD_CRC_SIZE (512)
#define CACHE_SIZE (16)                 /* number sectors in cache */
#define R1_LEN (48/8)
#define R2_LEN (136/8)
#define R3_LEN (48/8)
#define R6_LEN (48/8)
#define R7_LEN (48/8)
#define INIT_RETRIES (64)
#define RESP_TIME_I (64)
#define RESP_TIME_R (1024/2)
#define RESP_TIME_W (1024*4)

/* global variables and arrays */
uint16_t cardType;                		/* b0 = V3 MMC, b1 = V1 SD, b2 = V2 SD, b3 = HC/XC, b15 = funky read timing */
static uint8_t pkt[6];            		/* command packet */
static uint32_t num_sectors;
static uint32_t sd_speed;
static uint32_t sec_tags[CACHE_SIZE];
static uint8_t __attribute__((aligned(16))) sec_cache[CACHE_SIZE*512 + 8];
static uint8_t __attribute__((aligned(16))) sec_buf[520]; /* for uncached reads */
static uint8_t __attribute__((aligned(16))) resp_buf[32];
static uint8_t sd_csd[R2_LEN];
static int32_t respTime = RESP_TIME_R;

static const uint8_t __attribute__((aligned(16))) crc7_table[256] =
{
    0x00, 0x09, 0x12, 0x1b, 0x24, 0x2d, 0x36, 0x3f,
    0x48, 0x41, 0x5a, 0x53, 0x6c, 0x65, 0x7e, 0x77,
    0x19, 0x10, 0x0b, 0x02, 0x3d, 0x34, 0x2f, 0x26,
    0x51, 0x58, 0x43, 0x4a, 0x75, 0x7c, 0x67, 0x6e,
    0x32, 0x3b, 0x20, 0x29, 0x16, 0x1f, 0x04, 0x0d,
    0x7a, 0x73, 0x68, 0x61, 0x5e, 0x57, 0x4c, 0x45,
    0x2b, 0x22, 0x39, 0x30, 0x0f, 0x06, 0x1d, 0x14,
    0x63, 0x6a, 0x71, 0x78, 0x47, 0x4e, 0x55, 0x5c,
    0x64, 0x6d, 0x76, 0x7f, 0x40, 0x49, 0x52, 0x5b,
    0x2c, 0x25, 0x3e, 0x37, 0x08, 0x01, 0x1a, 0x13,
    0x7d, 0x74, 0x6f, 0x66, 0x59, 0x50, 0x4b, 0x42,
    0x35, 0x3c, 0x27, 0x2e, 0x11, 0x18, 0x03, 0x0a,
    0x56, 0x5f, 0x44, 0x4d, 0x72, 0x7b, 0x60, 0x69,
    0x1e, 0x17, 0x0c, 0x05, 0x3a, 0x33, 0x28, 0x21,
    0x4f, 0x46, 0x5d, 0x54, 0x6b, 0x62, 0x79, 0x70,
    0x07, 0x0e, 0x15, 0x1c, 0x23, 0x2a, 0x31, 0x38,
    0x41, 0x48, 0x53, 0x5a, 0x65, 0x6c, 0x77, 0x7e,
    0x09, 0x00, 0x1b, 0x12, 0x2d, 0x24, 0x3f, 0x36,
    0x58, 0x51, 0x4a, 0x43, 0x7c, 0x75, 0x6e, 0x67,
    0x10, 0x19, 0x02, 0x0b, 0x34, 0x3d, 0x26, 0x2f,
    0x73, 0x7a, 0x61, 0x68, 0x57, 0x5e, 0x45, 0x4c,
    0x3b, 0x32, 0x29, 0x20, 0x1f, 0x16, 0x0d, 0x04,
    0x6a, 0x63, 0x78, 0x71, 0x4e, 0x47, 0x5c, 0x55,
    0x22, 0x2b, 0x30, 0x39, 0x06, 0x0f, 0x14, 0x1d,
    0x25, 0x2c, 0x37, 0x3e, 0x01, 0x08, 0x13, 0x1a,
    0x6d, 0x64, 0x7f, 0x76, 0x49, 0x40, 0x5b, 0x52,
    0x3c, 0x35, 0x2e, 0x27, 0x18, 0x11, 0x0a, 0x03,
    0x74, 0x7d, 0x66, 0x6f, 0x50, 0x59, 0x42, 0x4b,
    0x17, 0x1e, 0x05, 0x0c, 0x33, 0x3a, 0x21, 0x28,
    0x5f, 0x56, 0x4d, 0x44, 0x7b, 0x72, 0x69, 0x60,
    0x0e, 0x07, 0x1c, 0x15, 0x2a, 0x23, 0x38, 0x31,
    0x46, 0x4f, 0x54, 0x5d, 0x62, 0x6b, 0x70, 0x79
};


static void (*wrMmcCmdBit)(uint32_t);
static void (*wrMmcCmdByte)(uint32_t);
static uint32_t (*rdMmcCmdBit)(void);
static uint32_t (*rdMmcCmdByte)(void);
static uint32_t (*rdMmcCmdByte7)(void);

static void (*wrMmcDatBit)(uint32_t);
static void (*wrMmcDatByte)(uint32_t);
static void (*wrMmcDatBit4)(uint32_t);
static void (*wrMmcDatByte4)(uint32_t);
static uint32_t (*rdMmcDatBit)(void);
static uint32_t (*rdMmcDatByte)(void);
static uint32_t (*rdMmcDatBit4)(void);
static uint32_t (*rdMmcDatByte4)(void);


/*-----------------------------------------------------------------------*/
/* N64 Myth Support                                                      */
/*-----------------------------------------------------------------------*/

static inline void sd_delay(int32_t cnt)
{
    for(int32_t ix=0; ix<cnt; ix++)
        asm("\tnop\n"::);
}

static void wrMythCmdBit( uint32_t bit )
{
    uint32_t data;
    vuint32_t *addr = (vuint32_t*)
       (( 0xB2000000 ) |       // GBA cart space
        ( 1<<24 ) |            // always 1
        ( 1<<19 ) |            // clk enable
        ( 1<<18 ) |            // cs remap
        ( 1<<17 ) |            // cmd remap
        ( 1<<16 ) |            // d1-d3 remap
        ( 1<<15 ) |            // d0 remap
        ( 1<<14 ) |            // cs value (active high)
        ( (bit&1)<<13 ) |      // cmd value
        ( 1<<12 ) |            // d3 value
        ( 1<<11 ) |            // d2 value
        ( 1<<10 ) |            // d1 value
        ( 1<<9 ));             // d0 value
    data = *addr;
    sd_delay(sd_speed);
}

static void wrMythCmdByte( uint32_t byte )
{
    wrMythCmdBit((byte>>7) & 1);
    wrMythCmdBit((byte>>6) & 1);
    wrMythCmdBit((byte>>5) & 1);
    wrMythCmdBit((byte>>4) & 1);
    wrMythCmdBit((byte>>3) & 1);
    wrMythCmdBit((byte>>2) & 1);
    wrMythCmdBit((byte>>1) & 1);
    wrMythCmdBit((byte) & 1);
}

static void wrMythDatBit( uint32_t bit )
{
    uint32_t data;
    vuint32_t *addr = (vuint32_t*)
       (( 0xB2000000 ) |       // GBA cart space
        ( 1<<24 ) |            // always 1
        ( 1<<19 ) |            // clk enable
        ( 1<<18 ) |            // cs remap
        ( 1<<17 ) |            // cmd remap
        ( 1<<16 ) |            // d1-d3 remap
        ( 1<<15 ) |            // d0 remap
        ( 1<<14 ) |            // cs value (active high)
        ( 1<<13 ) |            // cmd value
        ( 1<<12 ) |            // d3 value
        ( 1<<11 ) |            // d2 value
        ( 1<<10 ) |            // d1 value
        ( (bit&1)<<9 ));       // d0 value
    data = *addr;
    sd_delay(sd_speed);
}

static void wrMythDatByte( uint32_t byte )
{
    wrMythDatBit((byte>>7) & 1);
    wrMythDatBit((byte>>6) & 1);
    wrMythDatBit((byte>>5) & 1);
    wrMythDatBit((byte>>4) & 1);
    wrMythDatBit((byte>>3) & 1);
    wrMythDatBit((byte>>2) & 1);
    wrMythDatBit((byte>>1) & 1);
    wrMythDatBit((byte>>0) & 1);
}

static uint32_t rdMythCmdBit()
{
    uint32_t data;
    vuint32_t *addr = (vuint32_t*)
       (( 0xB2000000 ) |       // GBA cart space
        ( 1<<24 ) |            // always 1
        ( 1<<19 ) |            // clk enable
        ( 1<<18 ) |            // cs remap
        ( 1<<17 ) |            // cmd remap
        ( 1<<16 ) |            // d1-d3 remap
        ( 1<<15 ) |            // d0 remap
        ( 1<<14 ) |            // cs value (active high)
        ( 1<<13 ) |            // cmd value
        ( 1<<12 ) |            // d3
        ( 1<<11 ) |            // d2
        ( 1<<10 ) |            // d1
        ( 1<<9 )  |            // d0
        ( 1<<7 ));             // cmd input mode
    data = *addr;
    sd_delay(sd_speed);
    return (data>>12) & 1;
}

static uint32_t rdMythCmdByte()
{
    register uint32_t byte;

    byte = rdMmcCmdBit();
    byte = (byte << 1) | rdMythCmdBit();
    byte = (byte << 1) | rdMythCmdBit();
    byte = (byte << 1) | rdMythCmdBit();
    byte = (byte << 1) | rdMythCmdBit();
    byte = (byte << 1) | rdMythCmdBit();
    byte = (byte << 1) | rdMythCmdBit();
    byte = (byte << 1) | rdMythCmdBit();

    return byte;
}

static uint32_t rdMythCmdByte7()
{
    register uint32_t byte;

    byte = rdMmcCmdBit();
    byte = (byte << 1) | rdMmcCmdBit();
    byte = (byte << 1) | rdMmcCmdBit();
    byte = (byte << 1) | rdMmcCmdBit();
    byte = (byte << 1) | rdMmcCmdBit();
    byte = (byte << 1) | rdMmcCmdBit();
    byte = (byte << 1) | rdMmcCmdBit();

    return byte;
}

static uint32_t rdMythDatBit()
{
    uint32_t data;
    vuint32_t *addr = (vuint32_t*)
       (( 0xB2000000 ) |       // GBA cart space
        ( 1<<24 ) |            // always 1 (set in GBAC_LIO)
        ( 1<<19 ) |            // clk enable
        ( 1<<18 ) |            // cs remap
        ( 1<<17 ) |            // cmd remap
        ( 1<<16 ) |            // d1-d3 remap
        //( 1<<15 ) |          // d0 remap
        ( 1<<14 ) |            // cs value (active high)
        ( 1<<13 ) |            // cmd value
        ( 1<<12 ) |            // d3
        ( 1<<11 ) |            // d2
        ( 1<<10 ) |            // d1
        //( 1<<9 ) |           // d0
        //( 1<<6 ) |           // d3-d1 input (only needed in 4 bit mode)
        ( 1<<5 ));             // d0 input
    data = *addr;
    sd_delay(sd_speed);
    return (data>>8) & 1;
}

static uint32_t rdMythDatBit4()
{
    uint32_t data;
    vuint32_t *addr = (vuint32_t*)
       (( 0xB2000000 ) |       // GBA cart space
        ( 1<<24 ) |            // always 1 (set in GBAC_LIO)
        ( 1<<19 ) |            // clk enable
        ( 1<<18 ) |            // cs remap
        ( 1<<17 ) |            // cmd remap
        //( 1<<16 ) |          // d1-d3 remap
        //( 1<<15 ) |          // d0 remap
        ( 1<<14 ) |            // cs value (active high)
        ( 1<<13 ) |            // cmd value
        //( 1<<12 ) |          // d3
        //( 1<<11 ) |          // d2
        //( 1<<10 ) |          // d1
        //( 1<<9 ) |           // d0
        ( 1<<6 ) |             // d3-d1 input (only needed in 4 bit mode)
        ( 1<<5 ));             // d0 input
    data = *addr;
    sd_delay(sd_speed);
    return (data>>8) & 15;
}

static void wrMythDatBit4( uint32_t dat )
{
    uint32_t data;
    vuint32_t *addr = (vuint32_t*)
       (( 0xB2000000 ) |       // GBA cart space
        ( 1<<24 ) |            // always 1 (set in GBAC_LIO)
        ( 1<<19 ) |            // clk enable
        ( 1<<18 ) |            // cs remap
        ( 1<<17 ) |            // cmd remap
        ( 1<<16 ) |            // d1-d3 remap
        ( 1<<15 ) |            // d0 remap
        ( 1<<14 ) |            // cs value (active high)
        ( 1<<13 ) |            // cmd value
        ( (dat&15)<<9 ));      // d3-d0 value
    data = *addr;
    sd_delay(sd_speed);
}

static void wrMythDatByte4( uint32_t val )
{
    wrMythDatBit4((val >> 4) & 15);
    wrMythDatBit4(val & 15);
}

static uint32_t rdMythDatByte4()
{
    return ( (rdMythDatBit4() << 4) | rdMythDatBit4() );
}

static uint32_t rdMythDatByte()
{
    register uint32_t byte;

    byte = rdMythDatBit();
    byte = (byte << 1) | rdMythDatBit();
    byte = (byte << 1) | rdMythDatBit();
    byte = (byte << 1) | rdMythDatBit();
    byte = (byte << 1) | rdMythDatBit();
    byte = (byte << 1) | rdMythDatBit();
    byte = (byte << 1) | rdMythDatBit();

    return byte;
}

/*-----------------------------------------------------------------------*/
/* ED64 Support                                                          */
/*-----------------------------------------------------------------------*/

uint8_t evd_isDmaBusy() 
{
    uint16_t val;
    //volatile uint32_t i;
    sleep(1);
    if(dma_busy_callback != 0)dma_busy_callback();
    //regs_ptr[REG_STATE]++;
    val = regs_ptr[REG_STATUS];
    return (val >> ED_STATE_DMA_BUSY) & 1;
}

uint8_t evd_isDmaTimeout() 
{
    uint16_t val;
    //regs_ptr[REG_STATE]++;
    val = regs_ptr[REG_STATUS];

    return (val >> ED_STATE_DMA_TOUT) & 1;
}

uint8_t evd_isSpiBusy() 
{
    volatile uint16_t val;
    regs_ptr[REG_STATUS];
    val = regs_ptr[REG_STATUS];
    return (val >> ED_STATE_SPI) & 1;
}

uint8_t evd_SPI(uint8_t dat) 
{
    volatile uint8_t val;
    val = regs_ptr[0];
    regs_ptr[REG_SPI] = dat;
    while (evd_isSpiBusy());
    //osInvalICache((uint32_t*) & regs_ptr[REG_SPI], 1);
    val = regs_ptr[REG_SPI];
    return val;
}

void evd_spiSSOn() 
{
    volatile uint8_t val;
    if (sd_mode)return;
    spi_cfg &= ~(1 << SPI_CFG_SS);
    val = regs_ptr[0];
    regs_ptr[REG_SPI_CFG] = spi_cfg;
}

void evd_spiSSOff() 
{
    volatile uint8_t val;
    spi_cfg |= (1 << SPI_CFG_SS);
    val = regs_ptr[0];
    regs_ptr[REG_SPI_CFG] = spi_cfg;
}

void evd_enableSDMode() 
{
    sd_mode = 1;
}

void evd_enableSPIMode() 
{
    sd_mode = 0;
}

uint8_t evd_isSDMode() 
{
    return sd_mode;
}

void evd_SDcmdWriteMode(uint8_t bit1_mode) 
{
    volatile uint8_t val;
    if (!sd_mode)return;
    spi_cfg &= ~((1 << SPI_CFG_RD) | (1 << SPI_CFG_DAT));
    if (bit1_mode) {
        spi_cfg |= (1 << SPI_CFG_1BIT);
    } else {
        spi_cfg &= ~(1 << SPI_CFG_1BIT);
    }
    val = regs_ptr[0];
    regs_ptr[REG_SPI_CFG] = spi_cfg;
}

void evd_SDcmdReadMode(uint8_t bit1_mode) 
{
    volatile uint8_t val;
    if (!sd_mode)return;
    spi_cfg |= (1 << SPI_CFG_RD);
    spi_cfg &= ~(1 << SPI_CFG_DAT);
    if (bit1_mode) {
        spi_cfg |= (1 << SPI_CFG_1BIT);
    } else {
        spi_cfg &= ~(1 << SPI_CFG_1BIT);
    }
    val = regs_ptr[0];
    regs_ptr[REG_SPI_CFG] = spi_cfg;
}

void evd_SDdatWriteMode(uint8_t bit4_mode) 
{
    volatile uint8_t val;
    if (!sd_mode)return;
    spi_cfg &= ~(1 << SPI_CFG_RD);
    spi_cfg |= (1 << SPI_CFG_DAT);
    if (bit4_mode) {
        spi_cfg |= (1 << SPI_CFG_1BIT);
    } else {
        spi_cfg &= ~(1 << SPI_CFG_1BIT);
    }
    val = regs_ptr[0];
    regs_ptr[REG_SPI_CFG] = spi_cfg;
}

void evd_SDdatReadMode(uint8_t bit4_mode) 
{
    volatile uint8_t val;
    if (!sd_mode)return;
    spi_cfg |= (1 << SPI_CFG_RD) | (1 << SPI_CFG_DAT);
    if (bit4_mode) {
        spi_cfg |= (1 << SPI_CFG_1BIT);
    } else {
        spi_cfg &= ~(1 << SPI_CFG_1BIT);
    }
    val = regs_ptr[0];
    regs_ptr[REG_SPI_CFG] = spi_cfg;
}

void evd_setSpiSpeed(uint8_t speed) 
{
    volatile uint8_t val;
    spi_cfg &= ~3; //((1 << SPI_CFG_SPD0) | (1 << SPI_CFG_SPD1));
    spi_cfg |= speed & 3;
    val = regs_ptr[0];
    regs_ptr[REG_SPI_CFG] = spi_cfg;
}

uint8_t evd_mmcReadToCart(uint32_t cart_addr, uint32_t len) 
{
    volatile uint8_t val;
    cart_addr /= 2048;

    val = regs_ptr[0];
    regs_ptr[REG_DMA_LEN] = (len - 1);
    val = regs_ptr[0];
    regs_ptr[REG_DMA_RAM_ADDR] = cart_addr;
    val = regs_ptr[0];
    regs_ptr[REG_DMA_CFG] = DCFG_SD_TO_RAM;

    while (evd_isDmaBusy());
    if (evd_isDmaTimeout())return EVD_ERROR_MMC_TIMEOUT;

    return 0;
}

void evd_setCfgBit(uint8_t option, uint8_t state) 
{
    volatile uint8_t val;

    if (state)evd_cfg |= (1 << option);
    else
        evd_cfg &= ~(1 << option);

    val = regs_ptr[0];
    regs_ptr[REG_CFG] = evd_cfg;
    val = regs_ptr[0];
}

uint16_t evd_readReg(uint8_t reg) 
{
    volatile uint32_t tmp;
    
    tmp = regs_ptr[0];

    return regs_ptr[reg];
}

void evd_writeReg(uint8_t reg, uint16_t val) 
{
    volatile uint8_t tmp;
    tmp = regs_ptr[0];
    regs_ptr[reg] = val;
}

void evd_mmcSetDmaSwap(uint8_t state) 
{
    evd_setCfgBit(ED_CFG_SWAP, state);
}

void evd_writeMsg(uint8_t dat) 
{
    evd_writeReg(REG_MSG, dat);
}

uint8_t evd_readMsg() 
{
    return evd_readReg(REG_MSG);
}




/*-----------------------------------------------------------------------*/
/*                                                                       */
/*-----------------------------------------------------------------------*/

static uint8_t crc7(register uint8_t *buf)
{
    register uint8_t crc = 0;

    crc = crc7_table[(crc) ^ buf[0]];
    crc = crc7_table[(crc << 1) ^ buf[1]];
    crc = crc7_table[(crc << 1) ^ buf[2]];
    crc = crc7_table[(crc << 1) ^ buf[3]];
    crc = crc7_table[(crc << 1) ^ buf[4]];

    return crc;
}

static inline void sendMmcCmd( uint8_t cmd, uint32_t arg )
{
    pkt[0] = cmd | 0x40;                /* b7 = 0 => start bit, b6 = 1 => host command */
    pkt[1] = (uint8_t)(arg >> 24);
    pkt[2] = (uint8_t)(arg >> 16);
    pkt[3] = (uint8_t)(arg >> 8);
    pkt[4] = (uint8_t)(arg);
    pkt[5] = (crc7(pkt) << 1) | 1;      /* b0 = 1 => stop bit */
    wrMmcCmdByte( pkt[0] );
    wrMmcCmdByte( pkt[1] );
    wrMmcCmdByte( pkt[2] );
    wrMmcCmdByte( pkt[3] );
    wrMmcCmdByte( pkt[4] );
    wrMmcCmdByte( pkt[5] );
}

static int32_t recvMmcCmdResp( uint8_t *resp, uint32_t len, int32_t cflag )
{
    register uint32_t i;
    register uint8_t* r = resp;

    resp[0] = 0;
    for (i=0; i<respTime; i++)
    {
        /* wait on start bit */
        if(rdMmcCmdBit()==0)
        {
            *r++ = rdMmcCmdByte7();

            while(--len)
                *r++ = rdMmcCmdByte();

            if (cflag)
                wrMmcCmdByte(0xFF);     /* 8 cycles to complete the operation so clock can halt */

            respTime = RESP_TIME_R;
            return 1;
        }
    }

    respTime = RESP_TIME_R;
    return 0;
}

static int32_t sdReadSingleBlock( uint8_t *buf, uint32_t addr )
{
	register uint8_t* resp = resp_buf;

    sendMmcCmd(17, addr);               /* READ_SINGLE_BLOCK */
    if (!(cardType & 0x8000))
    {
        if (!recvMmcCmdResp(resp, R1_LEN, 0) || (resp[0] != 17))
            return 0;
    }

#if 0
    while ((rdMmcDatBit4()&1) != 0)
    {
        if (i-- <= 0)
        {
            return 0;               	/* timeout on start bit */
        }
    }

    for (i=0; i<SD_CRC_SIZE; i++)
        buf[i] = rdMmcDatByte4();

    /* Clock out CRC */
    for (i=0; i<8; i++)
        rdMmcDatByte4();

    /* Clock out end bit */
    rdMmcDatBit4();
#else
    asm_recv_sd_multi(buf,1);
#endif
    wrMmcCmdByte(0xFF);                 /* 8 cycles to complete the operation so clock can halt */

    return 1;
}

static inline int32_t sdReadStartMulti( uint32_t addr )
{
	register uint8_t* resp = resp_buf;

    sendMmcCmd(18, addr);               /* READ_MULTIPLE_BLOCK */
    if (!(cardType & 0x8000))
    {
        if (!recvMmcCmdResp(resp, R1_LEN, 0) || (resp[0] != 18))
            return 0;
    }

    return 1;
}

static int32_t sdReadMultiBlocks( BYTE *buff, BYTE count )
{
#if 0
    int32_t i, j, k;
    for (j=0, k=0; j<count; j++, k+=512)
    {
        i = 64*1024;
        while ((rdMmcDatBit4()&1) != 0)
            if (i-- <= 0)
            {
                return 0;           	/* timeout on start bit */
            }

        for (i=0; i<512; i+=4)
            *(uint32_t*)&buff[i+k] = (rdMmcDatByte4()<<24)|(rdMmcDatByte4()<<16)|(rdMmcDatByte4()<<8)|rdMmcDatByte4();

        /* Clock out CRC */
        rdMmcDatByte4();
        rdMmcDatByte4();
        rdMmcDatByte4();
        rdMmcDatByte4();
        rdMmcDatByte4();
        rdMmcDatByte4();
        rdMmcDatByte4();
        rdMmcDatByte4();

        /* Clock out end bit */
        rdMmcDatBit4();
    }
    return 1;
#else
    return asm_recv_sd_multi(buff, count);
#endif
}

static inline int32_t sdReadStopMulti( void )
{
	register uint8_t* resp = resp_buf;

    sendMmcCmd(12, 0);                  /* STOP_TRANSMISSION */
    if (!recvMmcCmdResp(resp, R1_LEN, 1) || (resp[0] != 12))
        return 0;

    return 1;
}

static void sdCrc16(uint8_t *p_crc, uint8_t *data, int32_t len)
{
    int32_t i;
    uint8_t nybble;

    unsigned long long poly = 0x0001000000100001LL;
    unsigned long long crc = 0;
    unsigned long long n_crc; 			/* This can probably be uint32_t */

    /* Load crc from array */
    for (i = 0; i < 8; i++)
    {
        crc <<= 8;
        crc |= p_crc[i];
    }

    for (i = 0; i < (len * 2); i++)
    {
        if (i & 1)
            nybble = (data[i >> 1] & 0x0F);
        else
            nybble = (data[i >> 1] >> 4);

        n_crc = (crc >> (15 * 4));
        crc <<= 4;
        if ((nybble ^ n_crc) & 1) crc ^= (poly << 0);
        if ((nybble ^ n_crc) & 2) crc ^= (poly << 1);
        if ((nybble ^ n_crc) & 4) crc ^= (poly << 2);
        if ((nybble ^ n_crc) & 8) crc ^= (poly << 3);
    }

    /* Output crc to array */
    for (i = 7; i >= 0; i--)
    {
        p_crc[i] = crc;
        crc >>= 8;
    }
}

static int32_t sendSdWriteBlock4( uint8_t *buf, uint8_t *crcbuf )
{
    int32_t i;

    wrMmcDatByte4(0xFF);                /* minimum 2 P bits */
    wrMmcDatBit4(0);                    /* write start bit */

    /* write out data and crc bytes */
    for (i=0; i<SD_CRC_SIZE; i++)
        wrMmcDatByte4( buf[i] );

    for (i=0; i<8; i++)
        wrMmcDatByte4( crcbuf[i] );

    wrMmcDatBit4(15);                   /* write end bit */

    rdMmcDatByte4();                    /* clock out two bits on D0 */

#if 1
    /*
     * spec says the CRC response from the card appears now...
     * start bit, three CRC response bits, end bit
     * this is followed by a busy response while the block is written
     * start bit, variable number of 0 bits, end bit
     */

    if( rdMmcDatBit4() & 1 )
    {
        return 0;						/* no start bit */
    }
    else
    {
        uint8_t crc_stat = 0;

        for(i=0; i<3; i++)
        {
            crc_stat <<= 1;
            crc_stat |= (rdMmcDatBit4() & 1);   /* read crc status on D0 */
        }
        rdMmcDatBit4();                 /* end bit */

        /* If CRC Status is not positive (010b) return error */
        if(crc_stat!=2)
        {
            return 0;					/* CRC error */
        }

        /* at this point the card is definately cooperating so wait for the start bit */
        while( rdMmcDatBit4() & 1 );

        /* Do not time out? Writes take an unpredictable amount of time to complete. */
        for(i=(1024*128);i>0;i--)
        {
            if((rdMmcDatBit4()&1) != 0 )
                break;
        }
        /* check for busy timeout */
        if(i==0)
        {
            return 0; /* busy timeout */
        }
    }
#else
    /*
     * ignore the spec - lets just clock out some bits and wait
     * for the busy to clear...it works well on many non-working cards
     */
    rdMmcDatByte4();                    /* clock out two bits on D0 */
    rdMmcDatByte4();                    /* clock out two bits on D0 */

    while(rdMmcDatBit4() & 1) ;         /* wait for DAT line to go low, indicating busy writing */
    while((rdMmcDatBit4()&1) == 0) ;    /* wait for DAT line high, indicating finished writing block */
#endif

    wrMmcCmdByte(0xFF);                 /* 8 cycles to complete the operation so clock can halt */

    return 1;
}

static int32_t sdWriteSingleBlock( uint8_t *buf, uint32_t addr )
{
    uint8_t resp[R1_LEN];
    uint8_t crcbuf[8];

    *(uint32_t*)&crcbuf[0]=0;
    *(uint32_t*)&crcbuf[4]=0;

    sdCrc16(crcbuf, buf, SD_CRC_SIZE);  /* Calculate CRC16 */

    sendMmcCmd( 24, addr );
    respTime = RESP_TIME_W;
    if (recvMmcCmdResp(resp, R1_LEN, 0) && (resp[0] == 24))
        return sendSdWriteBlock4(buf, crcbuf);

    return 0;
}

static int32_t sdInit(void)
{
    int32_t i;
    uint16_t rca;
    uint8_t resp[R2_LEN];         /* R2 is largest response */

    sd_speed = 4000;                    /* init with slow data rate */

    for (i=0; i<64; i++)
        wrMmcCmdBit(1);

    sendMmcCmd(0, 0);                   /* GO_IDLE_STATE */
    wrMmcCmdByte(0xFF);                 /* 8 cycles to complete the operation so clock can halt */

    sendMmcCmd(8, 0x1AA);               /* SEND_IF_COND */
    respTime = RESP_TIME_I;             /* init response time (extra short) */
    if (recvMmcCmdResp(resp, R7_LEN, 1))
    {
        if ((resp[0] == 8) && (resp[3] == 1) && (resp[4] == 0xAA))
            cardType |= CT_SD2;         /* V2 and/or HC card */
        else
            return 0;                   /* unusable */
    }

    for (i=0; i<INIT_RETRIES; i++)
    {
        sendMmcCmd(55, 0xFFFF);         /* APP_CMD */
        respTime = RESP_TIME_I;
        if (recvMmcCmdResp(resp, R1_LEN, 1))
        {
            if (resp[4] & 0x20)
            {
                sendMmcCmd(41, (cardType & CT_SD2) ? 0x40300000 : 0x00300000); /* SD_SEND_OP_COND */
                if (recvMmcCmdResp(resp, R3_LEN, 1) && (resp[0] == 0x3F) && (resp[1] & 0x80))
                {
                    if (resp[1] & 0x40)
                        cardType |= CT_BLOCK; /* HC/XC card */
                    if (!(resp[2] & 0x30))
                        return 0;       /* unusable */
                    break;
                }
            }
        }
        else
            return 0;                   /* no card */
    }
    if (i == INIT_RETRIES)
    {
        /* timed out */
        return 0;                       /* unusable */
    }

    sd_speed = 1000;                    /* continue with faster data rate */

    sendMmcCmd(2, 0xFFFFFFFF);          /* ALL_SEND_CID */
    if (!recvMmcCmdResp(resp, R2_LEN, 1) || (resp[0] != 0x3F))
        return 0;                       /* unusable */

    sendMmcCmd(3, 1);                   /* SEND_RELATIVE_ADDR */
    if (!recvMmcCmdResp(resp, R6_LEN, 1) || (resp[0] != 3))
        return 0;                       /* unusable */
    rca = (resp[1]<<8) | resp[2];

    sendMmcCmd(9, (((rca>>8)&0xFF)<<24) | ((rca&0xFF)<<16) | 0xFFFF); /* SEND_CSD */
    recvMmcCmdResp(sd_csd, R2_LEN, 1);

    sendMmcCmd(7, (((rca>>8)&0xFF)<<24) | ((rca&0xFF)<<16) | 0xFFFF); /* SELECT_DESELECT_CARD */
    if (!recvMmcCmdResp(resp, R1_LEN, 1) || (resp[0] != 7))
        return 0;                       /* unusable */

    sendMmcCmd(55, (((rca>>8)&0xFF)<<24) | ((rca&0xFF)<<16) | 0xFFFF); /* APP_CMD */
    if (!recvMmcCmdResp(resp, R1_LEN, 1) || !(resp[4] & 0x20))
        return 0;                       /* unusable */
    sendMmcCmd(6, 2);                   /* SET_BUS_WIDTH (to 4 bits) */
    if (!recvMmcCmdResp(resp, R1_LEN, 1) || (resp[0] != 6))
        return 0;                       /* unusable */

    return 1;
}

/*-----------------------------------------------------------------------*/
/* Initialize SD(HC) Card                                                */
/*-----------------------------------------------------------------------*/

DSTATUS MMC_disk_initialize (void)
{
    int32_t i;
    int32_t result = 0;

	if ((evd_cfg == 0) && (neo_firm == 0))
	{
		/* determine hardware and set function vectors */

	    evd_init();    					/* try to init EV64 */

	    if (evd_firm & 0xFF00)
	    {
            wrMmcCmdBit = wrEd64CmdBit;
            wrMmcCmdByte = wrEd64CmdByte;
            rdMmcCmdBit = rdEd64CmdBit;
            rdMmcCmdByte = rdEd64CmdByte;
            rdMmcCmdByte7 = rdEd64CmdByte7;

            wrMmcDatBit = wrEd64DatBit;
            wrMmcDatByte = wrEd64DatByte;
            wrMmcDatBit4 = wrEd64DatBit4;
            wrMmcDatByte4 = wrEd64DatByte4;
            rdMmcDatBit = rdEd64DatBit;
            rdMmcDatByte = rdEd64DatByte;
            rdMmcDatBit4 = rdEd64DatBit4;
            rdMmcDatByte4 = rdEd64DatByte4;
	    }
	    else
	    {
	        neo_enable_sd();
	        neo_firm = neo_get_cpld() & 15;

            wrMmcCmdBit = wrMythCmdBit;
            wrMmcCmdByte = wrMythCmdByte;
            rdMmcCmdBit = rdMythCmdBit;
            rdMmcCmdByte = rdMythCmdByte;
            rdMmcCmdByte7 = rdMythCmdByte7;

            wrMmcDatBit = wrMythDatBit;
            wrMmcDatByte = wrMythDatByte;
            wrMmcDatBit4 = wrMythDatBit4;
            wrMmcDatByte4 = wrMythDatByte4;
            rdMmcDatBit = rdMythDatBit;
            rdMmcDatByte = rdMythDatByte;
            rdMmcDatBit4 = rdMythDatBit4;
            rdMmcDatByte4 = rdMythDatByte4;
	    }
	}

    for (i=0; i<CACHE_SIZE; i++)
        sec_tags[i] = 0xFFFFFFFF;       /* invalidate cache entry */

    cardType &= 0x8000;                 /* keep funky flag */

    neo2_pre_sd();
    result = sdInit();
    neo2_post_sd();

    sd_speed = 0;                       /* data rate after init */
    if (!result)
    {
        cardType = 0xFFFF;
        return STA_NODISK;
    }

    if ((sd_csd[1] & 0xC0) == 0)
    {
        /*
         * CSD type 1 - version 1.x cards, and version 2.x standard capacity
         *
         *    C_Size      - 12 bits - [73:62]
         *    C_Size_Mult -  3 bits - [49:47]
         *    Read_Bl_Len -  4 bits - [83:80]
         *
         *    Capacity (bytes) = (C_Size + 1) * ( 2 ^ (C_Size_Mult + 2)) * (2 ^ Read_Bl_Len)
         */
        uint32_t C_Size      = ((sd_csd[7] & 0x03) << 10) | (sd_csd[8] << 2) | ((sd_csd[9] >>6) & 0x03);
        uint32_t C_Size_Mult = ((sd_csd[10] & 0x03) << 1) | ((sd_csd[11] >> 7) & 0x01);
        uint32_t Read_Bl_Len = sd_csd[6] & 0x0f;

        num_sectors = ((C_Size + 1) * (1 << (C_Size_Mult + 2)) * ((1 << Read_Bl_Len)) / 512);
    }
    else
    {
        /*
         * CSD type 2 - version 2.x high capacity
         *
         *    C_Size      - 22 bits - [69:48]
         *
         *    Capacity (bytes) = (C_Size + 1) * 1024 * 512
         */
        num_sectors = (((sd_csd[8] & 0x3F) << 16) | (sd_csd[9] << 8) | sd_csd[10]) * 1024;
    }

    return 0;
}

/*-----------------------------------------------------------------------*/
/* Return Disk Status                                                    */
/*-----------------------------------------------------------------------*/

DSTATUS MMC_disk_status (void)
{
    return (cardType != 0xFFFF) ? 0 : STA_NOINIT;
}

/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/

DRESULT MMC_disk_read (
    BYTE *buff,            				/* Data buffer to store read data */
    DWORD sector,        				/* Sector address (LBA) */
    BYTE count            				/* Number of sectors to read (1..255) */
)
{
    uint32_t ix;

    if (count == 1)
    {
        ix = sector & (CACHE_SIZE - 1); /* really simple hash */
        if (sec_tags[ix] != sector)
        {
            /* sector not in cache - fetch it */
            neo2_pre_sd();
            if (!sdReadSingleBlock(&sec_cache[ix << 9], sector << ((cardType & 1) ? 0 : 9)))
            {
                /* read failed, retry once */
                if (!sdReadSingleBlock(&sec_cache[ix << 9], sector << ((cardType & 1) ? 0 : 9)))
                {
                    /* read failed */
                    neo2_post_sd();
                    sec_tags[ix] = 0xFFFFFFFF;
                    return RES_ERROR;
                }
            }
            neo2_post_sd();
            sec_tags[ix] = sector;
        }

        neo_memcpy32(buff, &sec_cache[ix << 9], 512);
    }
    else
    {
        neo2_pre_sd();
        if (!sdReadStartMulti(sector << ((cardType & 1) ? 0 : 9)))
        {
            /* read failed, retry once */
            if (!sdReadStartMulti(sector << ((cardType & 1) ? 0 : 9)))
            {
                /* start multiple sector read failed */
                neo2_post_sd();
                return RES_ERROR;
            }
            if (!sdReadMultiBlocks(buff, count))
            {
                /* read failed */
                neo2_post_sd();
                return RES_ERROR;
            }
        }
        else if (!sdReadMultiBlocks(buff, count))
        {
            /* read failed, retry once */
            if (!sdReadStartMulti(sector << ((cardType & 1) ? 0 : 9)))
            {
                /* start multiple sector read failed */
                neo2_post_sd();
                return RES_ERROR;
            }
            if (!sdReadMultiBlocks(buff, count))
            {
                /* read failed */
                neo2_post_sd();
                return RES_ERROR;
            }
        }
        sdReadStopMulti();
        neo2_post_sd();
    }

    return RES_OK;
}

DRESULT MMC_disk_read_multi(BYTE* buff,DWORD sector,UINT count)
{
	neo2_pre_sd();
    {
		sector <<= ((cardType & 1) ? 0 : 9);

        if (!sdReadStartMulti(sector))
        {
            /* read failed, retry once */
            if (!sdReadStartMulti(sector))
            {
                /* start multiple sector read failed */
                neo2_post_sd();
                return RES_ERROR;
            }
            if (!sdReadMultiBlocks(buff,count))
            {
                /* read failed */
                neo2_post_sd();
                return RES_ERROR;
            }

			sdReadStopMulti();
			neo2_post_sd();
			return RES_OK;
        }

		if (!sdReadMultiBlocks(buff,count))
        {
            /* read failed, retry once */
            if (!sdReadStartMulti(sector))
            {
                /* start multiple sector read failed */
                neo2_post_sd();
                return RES_ERROR;
            }
            if (!sdReadMultiBlocks(buff,count))
            {
                /* read failed */
                neo2_post_sd();
                return RES_ERROR;
            }
        }
   		 sdReadStopMulti();
    }

    neo2_post_sd();
    return RES_OK;
}

/*-----------------------------------------------------------------------*/
/* Write Sector(s)                                                       */
/*-----------------------------------------------------------------------*/

int32_t disk_io_force_wdl = 0;

DRESULT MMC_disk_write (
    const BYTE *buff,    				/* Data to be written */
    DWORD sector,        				/* Sector address (LBA) */
    BYTE count           				/* Number of sectors to write (1..255) */
)
{
    uint32_t ix, iy;

    sd_speed = disk_io_force_wdl + 1;   /* no fast access on writes */
    for (ix=0; ix<count; ix++)
    {
        iy = (sector + ix) & (CACHE_SIZE - 1); /* really simple hash */
        if (sec_tags[iy] == (sector + ix))
            sec_tags[iy] = 0xFFFFFFFF;

        neo_memcpy32(sec_buf, (void *)((uint32_t)buff + (ix << 9) ), 512);
        neo2_pre_sd();
        if (!sdWriteSingleBlock(sec_buf, (sector + ix) << ((cardType & 1) ? 0 : 9)))
        {
            /* write failed, retry */
            if (!sdWriteSingleBlock(sec_buf, (sector + ix) << ((cardType & 1) ? 0 : 9)))
            {
				/* retry */
				if (!sdWriteSingleBlock(sec_buf, (sector + ix) << ((cardType & 1) ? 0 : 9)))
				{
		            neo2_post_sd();
		            sd_speed = 0;
		            return RES_ERROR;
				}
            }
        }
        neo2_post_sd();
    }

    sd_speed = 0;                       /* allow fast access on reads */
    return RES_OK;
}

/*-----------------------------------------------------------------------*/
/* Miscellaneous Functions                                               */
/*-----------------------------------------------------------------------*/

DRESULT MMC_disk_ioctl (
    BYTE ctrl,        					/* Control code */
    void *buff        					/* Buffer to send/receive control data */
)
{
    switch (ctrl)
    {
        case GET_SECTOR_COUNT:
            *(uint32_t *)buff = num_sectors;
            return num_sectors ? RES_OK : RES_ERROR;
        case GET_SECTOR_SIZE:
            *(uint16_t *)buff = 512;
            return RES_OK;
        case GET_BLOCK_SIZE:
            *(uint32_t *)buff = 1;
            return RES_OK;
        case CTRL_SYNC:
            return RES_OK;
        case CTRL_POWER:
        case CTRL_LOCK:
        case CTRL_EJECT:
        case MMC_GET_TYPE:
        case MMC_GET_CSD:
        case MMC_GET_CID:
        case MMC_GET_OCR:
        case MMC_GET_SDSTAT:
        default:
            return RES_PARERR;
    }
}

/* 31-25: Year(0-127 org.1980), 24-21: Month(1-12), 20-16: Day(1-31) */
/* 15-11: Hour(0-23), 10-5: Minute(0-59), 4-0: Second(0-29 *2) */
DWORD get_fattime (void)
{
    return 0;
}

/*-----------------------------------------------------------------------*/
/* Initialize a Drive                                                    */
/*-----------------------------------------------------------------------*/

DSTATUS disk_initialize (
	BYTE drv							/* Physical drive number (0..) */
)
{
	if ( drv == MMC )
	{
		int32_t result = MMC_disk_initialize();
		return result ? STA_NOINIT : 0;
	}

	return STA_NOINIT;
}

/*-----------------------------------------------------------------------*/
/* Get Disk Status                                                       */
/*-----------------------------------------------------------------------*/

DSTATUS disk_status (
	BYTE drv							/* Physical drive number (0..) */
)
{
	if ( drv == MMC )
	{
		int32_t result = MMC_disk_status();
		return result ? STA_NODISK : 0;
	}

	return STA_NOINIT;
}

/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/

DRESULT disk_read (
	BYTE drv,							/* Physical drive number (0..) */
	BYTE *buff,							/* Data buffer to store read data */
	DWORD sector,						/* Sector address (LBA) */
	BYTE count							/* Number of sectors to read (1..128) */
)
{
	if ( drv == MMC )
	{
		int32_t result = MMC_disk_read(buff, sector, count);
		return result ? RES_ERROR : RES_OK;
	}

	return RES_PARERR;
}

/*-----------------------------------------------------------------------*/
/* Write Sector(s)                                                       */
/*-----------------------------------------------------------------------*/

#if _USE_WRITE
DRESULT disk_write (
	BYTE drv,							/* Physical drive number (0..) */
	const BYTE *buff,					/* Data to be written */
	DWORD sector,						/* Sector address (LBA) */
	BYTE count							/* Number of sectors to write (1..128) */
)
{
	if ( drv == MMC )
	{
		int32_t result = MMC_disk_write(buff, sector, count);
		return result ? RES_ERROR : RES_OK;
	}

	return RES_PARERR;
}
#endif

/*-----------------------------------------------------------------------*/
/* Miscellaneous Functions                                               */
/*-----------------------------------------------------------------------*/

#if _USE_IOCTL
DRESULT disk_ioctl (
	BYTE drv,							/* Physical drive number (0..) */
	BYTE ctrl,							/* Control code */
	void *buff							/* Buffer to send/receive control data */
)
{
	if ( drv == MMC )
	{
		int32_t result = MMC_disk_ioctl(ctrl, buff);
		return result ? RES_ERROR : RES_OK;
	}

	return RES_PARERR;
}
#endif
