/**
 * @file sdfs.c
 * @brief SDCardFS
 * @ingroup sdfs
 */
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "libdragon.h"
#include "system.h"
#include "diskio.h"
#include "ff.h"

/**
 * @defgroup sdfs SDCardFS
 * @brief SDCardFS filesystem implementation and newlib hooks.
 *
 * SDCardFS uses libff to read/write FAT formatted SD and SDHC cards. The disk
 * interface currently supports the N64 NeoMyth and the Everdrive64 carts.
 *
 * SDFS files have a maximum size of 4 GBytes.  Directory and file limits are
 * determined by libff with the exception of a maximum path length of 1024 chars,
 * and individual object names (file or directory) can only be a maximum of 255
 * chars.
 *
 * When SDFS is initialized, it will register itself with newlib using 'sdc:/' as a prefix.
 * Files can be accessed either with standard POSIX functions and the 'sdc:/' prefix or
 * with libff API calls and no prefix. Files opened with the POSIX function MUST be closed
 * with the POSIX function or memory will be lost.
 *
 * The SD card is initialized and mounted when SDFS is initialized, and when findfirst
 * is called on the root. This allows for hot-swapping SD cards.
 * @{
 */

/** @brief Directory to walk */
static DIR directory;

/** @brief Global FAT filesystem */
static FATFS filesystem;

/** @brief Type of SD card */
extern uint16_t cardType;


/** @brief copy char string to wide char string */
static void wstrcpy(TCHAR *dst, char *src)
{
    int ix = 0;
    while (1)
    {
        dst[ix] = (TCHAR)src[ix];
        if ( dst[ix] == (TCHAR)0 )
            break;
        ix++;
    }
}

/** @brief copy char string to wide char string with maximum length */
static void wstrncpy(TCHAR *dst, char *src, int len)
{
    int ix = 0;
    while (ix < len)
    {
        dst[ix] = (TCHAR)src[ix];
        if ( dst[ix] == (TCHAR)0 )
            break;
        ix++;
    }
}

/** @brief copy wide char string to char string */
static void cstrcpy(char *dst, TCHAR *src)
{
    int ix = 0;
    while (1)
    {
        dst[ix] = src[ix] & 0x00FF;
        if ( dst[ix] == 0 )
            break;
        ix++;
    }
}

/** @brief try to mount SD card */
static void mount_sd(void)
{
    TCHAR wtemp[4];

    /* unmount SD card */
    f_mount(MMC, NULL);

    /* init SD card */
    cardType = 0x0000;
    if ( disk_initialize(MMC) != 0 )
    {
        cardType = 0xFFFF;          /* card not mounted */
        return;
    }

    /* mount the card */
    if ( f_mount(MMC, &filesystem) != FR_OK )
    {
        cardType = 0xFFFF;          /* card not mounted */
        return;
    }
    f_chdrive(MMC);                 /* make MMC current drive */

    /* now check for funky timing by trying to opendir on the root */
    memset(&directory, 0, sizeof(DIR));
    wstrcpy(wtemp, "/");
    if ( f_opendir(&directory, wtemp) )
    {
        /* failed, try funky timing */
        cardType = 0x8000;          /* try funky read timing */
        if ( disk_initialize(MMC) != 0 )
        {
            cardType = 0xFFFF;      /* card not mounted */
            return;
        }
        if ( f_opendir(&directory, wtemp) )
        {
            /* failed */
            cardType = 0xFFFF;      /* card not mounted */
            return;
        }
    }
}

/**
 * @brief Newlib-compatible open
 *
 * @param[in] name
 *            Absolute path of the file to open
 * @param[in] flags
 *            POSIX file flags
 *
 * @return A newlib-compatible file handle.
 */
static void *__open( char *name, int flags )
{
    BYTE mode;
    static TCHAR wtemp[1024 + 256];
    FIL *file = malloc(sizeof(FIL));
    if ( file == NULL )
        return NULL;

    /* Always want a consistent interface */
    wstrcpy(wtemp, "/");
    f_chdir(wtemp);

    /* make arguments for f_open */
    wstrncpy(wtemp, name, 1024 + 255);
    wtemp[1024 + 255] = (TCHAR)0;
    if ( (flags & 3) == O_RDONLY )
        mode = FA_READ;
    else if ( (flags & 3) == O_WRONLY )
        mode = FA_WRITE;
    else if ( (flags & 3) == O_RDWR )
        mode = FA_READ | FA_WRITE;
    if ( flags & O_CREAT )
        mode |= FA_OPEN_ALWAYS;

    if ( f_open(file, wtemp, flags) == FR_OK )
    {
        if ( flags & O_APPEND )
            f_lseek(file, f_size(file));
        else if ( flags & O_TRUNC )
            f_truncate(file);

        return (void *)file;
    }

    free(file);
    return NULL;
}

/**
 * @brief Newlib-compatible fstat
 *
 * @param[in]  file
 *             File pointer as returned by #__open
 * @param[out] st
 *             Stat structure to populate
 *
 * @return 0.
 */
static int __fstat( void *file, struct stat *st )
{
    st->st_dev = 0;
    st->st_ino = 0;
    st->st_mode = S_IFREG;
    st->st_nlink = 1;
    st->st_uid = 0;
    st->st_gid = 0;
    st->st_rdev = 0;
    st->st_size = f_size((FIL*)file);
    st->st_atime = 0;
    st->st_mtime = 0;
    st->st_ctime = 0;
    st->st_blksize = 0;
    st->st_blocks = 0;
    //st->st_attr = S_IAREAD | S_IAREAD;

    return 0;
}

/**
 * @brief Newlib-compatible lseek
 *
 * @param[in] file
 *            File pointer as returned by #__open
 * @param[in] ptr
 *            Offset based on dir
 * @param[in] dir
 *            A direction to seek from.  Either #SEEK_SET, #SEEK_CUR or #SEEK_END
 *
 * @return The new position in the file after the seek.
 */
static int __lseek( void *file, int ptr, int dir )
{
    DWORD offs = f_tell((FIL*)file);
    switch (dir)
    {
        case SEEK_SET:
            offs = ptr;
            break;
        case SEEK_CUR:
            offs += ptr;
            break;
        case SEEK_END:
            offs = f_size((FIL*)file) + ptr;
            break;
    }
    if ( offs > f_size((FIL*)file) )
        offs = f_size((FIL*)file);

    f_lseek((FIL*)file, offs);

    return f_tell((FIL*)file);
}

/**
 * @brief Newlib-compatible read
 *
 * @param[in]  file
 *             File pointer as returned by #__open
 * @param[out] ptr
 *             Pointer to buffer to read to
 * @param[in]  len
 *             Length in bytes to read
 *
 * @return The actual amount of data read.
 */
static int __read( void *file, uint8_t *ptr, int len )
{
    UINT res;

    if ( f_read((FIL*)file, ptr, len, &res) == FR_OK )
        return res;

    return 0;
}

/**
 * @brief Newlib-compatible write
 *
 * @param[in]  file
 *             File pointer as returned by #__open
 * @param[out] ptr
 *             Pointer to buffer to write from
 * @param[in]  len
 *             Length in bytes to write
 *
 * @return The actual amount of data written.
 */
static int __write( void *file, uint8_t *ptr, int len )
{
    UINT res;

    if ( f_write((FIL*)file, ptr, len, &res) == FR_OK )
        return res;

    return 0;
}

/**
 * @brief Newlib-compatible close
 *
 * @param[in] file
 *            File pointer as returned by #__open
 *
 * @return 0 on success or a negative error otherwise.
 */
static int __close( void *file )
{
    if ( f_close((FIL*)file) == FR_OK )
    {
        free(file);
        return 0;
    }

    return -1;
}

/**
 * @brief Newlib-compatible findfirst
 *
 * @param[in]  path
 *             Absolute path of the directory to walk
 * @param[out] dir
 *             Directory structure to populate with information on the first entry found
 *
 * @return 0 on success or a negative value on failure.
 */
static int __findfirst( char *path, dir_t *dir )
{
    static TCHAR wtemp[1024];
    static TCHAR lfnbuf[256];
    FILINFO fno;

    if( !path || !dir )
        return -1;

    if ( !strcmp(path, "/") )
    {
        /* calling findfirst on the root will reinit the SD card to handle media changes */
        mount_sd();
    }

    memset(&directory, 0, sizeof(DIR));
    wstrncpy(wtemp, path, 1023);
    wtemp[1023] = 0;
    if ( f_opendir(&directory, wtemp) != FR_OK )
        return -1;

    memset(&fno, 0, sizeof(FILINFO));
    fno.lfname = (TCHAR*)lfnbuf;
    fno.lfsize = 255;
    fno.lfname[0] = (TCHAR)0;

    if ( f_readdir(&directory, &fno) != FR_OK )
        return -1; /* error */

    if ( !fno.fname[0] )
        return -1; /* no entries */

    if ( fno.fattrib & AM_DIR )
        dir->d_type = DT_DIR;
    else
        dir->d_type = DT_REG;

    if ( fno.lfname[0] )
        cstrcpy(dir->d_name, fno.lfname);
    else
        strcpy(dir->d_name, fno.fname);

    /* Success */
    return 0;
}

/**
 * @brief Newlib-compatible findnext
 *
 * @param[out] dir
 *              Directory structure to populate with information on the next entry found
 *
 * @return 0 on success or a negative value on failure.
 */
static int __findnext( dir_t *dir )
{
    static TCHAR lfnbuf[256];
    FILINFO fno;

    if( !dir )
        return -1;

    memset(&fno, 0, sizeof(FILINFO));
    fno.lfname = (TCHAR*)lfnbuf;
    fno.lfsize = 255;
    fno.lfname[0] = (TCHAR)0;

    if ( f_readdir(&directory, &fno) != FR_OK )
        return -1; /* error */

    if ( !fno.fname[0] )
        return -1; /* no more entries */

    if ( fno.fattrib & AM_DIR )
        dir->d_type = DT_DIR;
    else
        dir->d_type = DT_REG;

    if(fno.lfname[0])
        cstrcpy(dir->d_name, fno.lfname);
    else
        strcpy(dir->d_name, fno.fname);

    /* Success */
    return 0;
}

/**
 * @brief Structure used for hooking DragonFS into newlib
 *
 * The following section of code is for bridging into newlib's filesystem hooks
 * to allow posix access to DragonFS filesystem.
 */
static filesystem_t sdcard_fs = {
    __open,
    __fstat,
    __lseek,
    __read,
    __write,
    __close,
    0,
    __findfirst,
    __findnext
};

/**
 * @brief Initialize the filesystem.
 *
 * This function will initialize the filesystem to read/write from/to an SD or
 * SDHC card.  This function will also register SDCardFS with newlib so that
 * standard POSIX file operations work with SDCardFS.
 *
 * @return 0 on success, or -1 otherwise.
 */
int sdfs_init(void)
{
    /* attempt to init SD interface and mount card */
    memset(&filesystem, 0, sizeof(FATFS));
    mount_sd();

    /*
     * With current carts, it makes no sense to fail just because an SD card
     * isn't found as one could be inserted later. So always register the
     * filesystem and return success.
     */

    /* push our filesystem into newlib */
    attach_filesystem( "sdc:/", &sdcard_fs );

    return 0;
}


/** @} */
